#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <complex>
#include <sys/time.h>
#include <unistd.h>

#include <cuda_runtime.h>
#include <cublas.h>
#include <cublas_v2.h>
#include "device_launch_parameters.h"
#include <cusolverDn.h>
#include <cuda_runtime_api.h>

using namespace std;

void gpuAssert(cudaError_t code, char *file, int line, bool abort=true)
{
   if (code != cudaSuccess)
   {
      fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
      if (abort) { exit(code); }
   }
}

void gpuErrchk(cudaError_t ans)
{
  gpuAssert((ans), __FILE__, __LINE__);
}

static void fprintf_matrix(complex<float>* A, int m, int n, int lda, const char* name)
{
  fstream outfile;
  outfile.open(name, ios::out);
  outfile.setf(std::ios_base::left, std::ios_base::basefield);

  for (int iRow = 0; iRow < m; iRow++)
  {
    for (int jCol = 0; jCol < n; jCol++)
      outfile << A[jCol * lda + iRow].real() << " " << A[jCol * lda + iRow].imag() << " ";
    outfile << endl;
  }

  outfile.close();
}

static void fprintf_matrix(float* A, int m, int n, int lda, const char* name)
{
  fstream outfile;
  outfile.open(name, ios::out);
  outfile.setf(std::ios_base::left, std::ios_base::basefield);

  for (int iRow = 0; iRow < m; iRow++)
  {
    for (int jCol = 0; jCol < n; jCol++)
      outfile << A[jCol * lda + iRow] << " ";
    outfile << endl;
  }

  outfile.close();
}

static void _printStatus(cusolverStatus_t status)
{
  cout << "status = " << status << endl;

  switch(status)
  {
      case CUSOLVER_STATUS_SUCCESS:
        cout << "success" << endl;
        break;
      case CUSOLVER_STATUS_NOT_INITIALIZED :
        cout << "Library cuSolver not initialized correctly" << endl;
        break;
      case CUSOLVER_STATUS_INVALID_VALUE:
        cout << "Invalid parameters passed" << endl;
        break;
      case CUSOLVER_STATUS_INTERNAL_ERROR:
        cout << "Internal operation failed" << endl;
        break;
      case CUSOLVER_STATUS_EXECUTION_FAILED:
        cout << "Execution failed" << endl;
        break;
  }
}

static void _svdCUDA
(
  cusolverDnHandle_t cuHandle,
  int m,
  int n,
  int ldA,
  cuComplex* d_A,
  int ldUsvd,
  int ldVsvdH,
  cuComplex* d_Usvd,
  float* d_Ssvd,
  cuComplex* d_VsvdH
)
{
  cusolverStatus_t status;

  int workSize = 0;
  status = cusolverDnCgesvd_bufferSize(cuHandle, m, n, &workSize);
  if (status != CUSOLVER_STATUS_SUCCESS )
    cout << "Initialization of cuSolver failed." << endl;

  int *devInfo;
  gpuErrchk(cudaMalloc(&devInfo,  sizeof(int)));
  cuComplex* work;
  gpuErrchk(cudaMalloc(&work, workSize * sizeof(cuComplex)));
  float* rwork;
  gpuErrchk(cudaMalloc(&rwork, workSize * sizeof(cuComplex)));

  status = cusolverDnCgesvd(cuHandle, 'A', 'A', m, n, d_A, ldA, d_Ssvd, d_Usvd, ldUsvd, d_VsvdH, ldVsvdH, work, workSize, rwork, devInfo);
  cudaDeviceSynchronize();
  _printStatus(status);

#if 0
  {
    complex<float>* h_Usvd = (complex<float> *)malloc(ldUsvd * m * sizeof(complex<float>));
    gpuErrchk(cudaMemcpy(h_Usvd, d_Usvd, ldUsvd * m * sizeof(cuComplex), cudaMemcpyDeviceToHost));
    fprintf_matrix(h_Usvd, m, m, m, "Usvd.dat");

    int minMN = min(m,n);
    float* h_Ssvd = (float *)malloc(minMN * sizeof(float));
    gpuErrchk(cudaMemcpy(h_Ssvd, d_Ssvd, minMN * sizeof(float), cudaMemcpyDeviceToHost));
    fprintf_matrix(h_Ssvd, minMN, 1, minMN, "Ssvd.dat");

    complex<float>* h_VsvdH = (complex<float> *)malloc(ldVsvdH * n * sizeof(complex<float>));
    gpuErrchk(cudaMemcpy(h_VsvdH, d_VsvdH, ldVsvdH * n * sizeof(cuComplex), cudaMemcpyDeviceToHost));
    fprintf_matrix(h_VsvdH, n, n, n, "VsvdH.dat");
  }
#endif
}

static void _setMatrix(int m, int n, complex<float>* pA)
{
  srand(static_cast<unsigned>(time(0)));

  for (int jCol = 0; jCol < n; jCol++)
    for (int iRow = 0; iRow < m; iRow++)
    {
      float real = rand() % 100;
      float imag = rand() % 100;
      pA[m * jCol + iRow] = complex<float>(real,imag);
    }

#if 0
  fprintf_matrix(pA, m, n, m, "A.dat");
#endif
}

int main()
{
  cusolverStatus_t statusCuSolver;
  cusolverDnHandle_t cuHandle;
  cusolverDnCreate(&cuHandle);

  int m = 1856;
  int n = 1849;
  int ldA = m;
  int minMN = min(m,n);
  complex<float>* h_A = (complex<float> *)malloc(m * n * sizeof(complex<float>));
  _setMatrix(m, n, h_A);

  cuComplex* d_A;
  gpuErrchk(cudaMalloc(&d_A, m * n * sizeof(cuComplex)));
  gpuErrchk(cudaMemcpy(d_A, h_A, m * n * sizeof(cuComplex), cudaMemcpyHostToDevice));

  int ldUt = ldA;
  cuComplex* d_Ut;
  gpuErrchk(cudaMalloc(&d_Ut, m * m * sizeof(cuComplex)));
  int ldVtH = n;
  cuComplex* d_VtH;
  gpuErrchk(cudaMalloc(&d_VtH, n * n * sizeof(cuComplex)));
  float* d_St;
  gpuErrchk(cudaMalloc(&d_St, minMN * sizeof(float)));

  struct timeval start, end;
  gettimeofday(&start, NULL);

  _svdCUDA(cuHandle, m, n, ldA, d_A, ldUt, ldVtH, d_Ut, d_St, d_VtH);

  gettimeofday(&end, NULL);
  double t = end.tv_sec + (end.tv_usec / 1000000.0) -
   (start.tv_sec + (start.tv_usec / 1000000.0));
  cout << "SVD time = " << t << endl;

  cusolverDnDestroy(cuHandle);

  return 0;
}
